from EasyNN.loss.abc import Loss
from EasyNN.loss.log_likelihood import LogLikelihood
from EasyNN.loss.mean_square_error import MeanSquareError
from EasyNN.loss.negative_likelihood import NegativeLikelihood
